/*
 * Copyright (c) 2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 ** velodyneExecTaskCodels.c
 **
 ** Codels called by execution task velodyneExecTask
 **
 **/

/* System headers */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

/* External headers */
#include <libimages3d.h>
#include <t3d/t3d.h>

/* Genom headers */
#include <portLib.h>

/* External Genom headers */
#include <libeuler/pomEuler.h>

/* Internal headers */
#include "server/velodyneHeader.h"
#include "velodyne3DImage.h"
#include "velodyneBuffer.h"

extern VELODYNE_INTERFACE* device;
extern velodyneReadTask	task;
extern velodyne3DImageParams imageParams;

static POSTER_ID	sensorPosPosterId;

/* Internal data */
static velodyneTarget		target;

/* Produced posters */
extern velodyne3DImage* threeDImage;
static GENPOS_TRAJ_POINTS*	targetAbsolutePosition;
static GENPOS_TRAJ_POINTS*	targetRelativePosition;

/* Function prototypes */
extern ACTIVITY_EVENT velodyneGrabOneShot(velodyneBuffer*, velodyneOneShotParams*, velodyne3DImage*, velodyne3DImageParams*, velodyneTargetDetectionParams*, velodyneTargetParams*, velodyneTarget*);
extern ACTIVITY_EVENT velodyne_analyzeTarget (velodyneTargetParams*, velodyneTargetDetectionParams*, velodyneTarget*);
extern ACTIVITY_EVENT velodyne_eraseTargetPoints (velodyne3DImage*, velodyneTarget*);

/*------------------------------------------------------------------------
 *
 * velodyneExecTaskStart  --  Initialization codel
 *
 * Description:
 *
 * Returns:    OK or ERROR
 */

STATUS
velodyneExecTaskStart(int *report)
{
  /* Sets all the request parameters to their default value */

  /* initParams */
  SDI_F->initParams.mode = VELODYNE_DEFAULT_MODE;
  SDI_F->initParams.model = VELODYNE_DEFAULT_MODEL;
  SDI_F->initParams.usePom = VELODYNE_DEFAULT_USE_POM;
  SDI_F->initParams.unused = VELODYNE_DEFAULT_UNUSED;
  strncpy (SDI_F->initParams.sensorPosPoster,
           VELODYNE_DEFAULT_SENSOR_POS_POSTER,
           VELODYNE_MAX_STRLEN);
  strncpy (SDI_F->initParams.inputFileName, VELODYNE_DEFAULT_INPUT_FILE_NAME,
           VELODYNE_MAX_STRLEN);
  SDI_F->initParams.enableOrganization = VELODYNE_DEFAULT_ENABLE_ORGANIZATION;
  SDI_F->initParams.discardAngularSector
      = VELODYNE_DEFAULT_DISCARD_ANGULAR_SECTOR;
  SDI_F->initParams.firstDiscardedAngle
      = VELODYNE_DEFAULT_FIRST_DISCARDED_ANGLE;
  SDI_F->initParams.lastDiscardedAngle = VELODYNE_DEFAULT_LAST_DISCARDED_ANGLE;

  /* Target detection params */
  SDI_F->targetDetectionParams.detectionOn = VELODYNE_DEFAULT_DETECTION_MODE;
  SDI_F->targetDetectionParams.trackingOn = VELODYNE_DEFAULT_TRACKING_MODE;
  SDI_F->targetDetectionParams.verbosity = VELODYNE_DEFAULT_VERBOSITY;
  SDI_F->targetDetectionParams.eraseTargetPoints = VELODYNE_DEFAULT_ERASE_TARGET_POINTS;

  /* Target params */
  SDI_F->targetParams.maxDistance = VELODYNE_DEFAULT_MAX_TARGET_DISTANCE;
  SDI_F->targetParams.zMin = VELODYNE_DEFAULT_TARGET_Z_MIN;
  SDI_F->targetParams.minReflectance = VELODYNE_DEFAULT_MIN_REFLECTANCE;
  SDI_F->targetParams.maxReflectance = VELODYNE_DEFAULT_MAX_REFLECTANCE;
  SDI_F->targetParams.radius = VELODYNE_DEFAULT_TARGET_RADIUS;
  SDI_F->targetParams.nbRequestedPoints = VELODYNE_DEFAULT_NB_REQUESTED_POINTS;
  SDI_F->targetParams.minWidth = VELODYNE_DEFAULT_TARGET_MIN_WIDTH;
  SDI_F->targetParams.maxWidth = VELODYNE_DEFAULT_TARGET_MAX_WIDTH;
  SDI_F->targetParams.minHeight = VELODYNE_DEFAULT_TARGET_MIN_HEIGHT;
  SDI_F->targetParams.maxHeight = VELODYNE_DEFAULT_TARGET_MAX_HEIGHT;

  /* saveParams */
  strncpy (SDI_F->saveParams.baseName, VELODYNE_DEFAULT_BASE_NAME,
           VELODYNE_MAX_STRLEN);
  SDI_F->saveParams.num = VELODYNE_DEFAULT_NUM;

  /* saveIm3DParams */
  strncpy (SDI_F->saveIm3DParams.baseName, VELODYNE_DEFAULT_BASE_NAME_IM3D,
           VELODYNE_MAX_STRLEN);
  SDI_F->saveIm3DParams.num = VELODYNE_DEFAULT_NUM;

  SDI_F->running = 0;

  /* Private structures */
  threeDImage = malloc (sizeof (velodyne3DImage));
  if (threeDImage == NULL) {
    fprintf (stderr, "velodyne: can't allocate 3D data structure\n");
    return ERROR;
  }
  targetAbsolutePosition = malloc (sizeof (GENPOS_TRAJ_POINTS));
  if (targetAbsolutePosition == NULL) {
    fprintf (stderr, "velodyne: can't allocate target structure\n");
    return ERROR;
  }
  targetRelativePosition = malloc (sizeof (GENPOS_TRAJ_POINTS));
  if (targetRelativePosition == NULL) {
    fprintf (stderr, "velodyne: can't allocate target structure\n");
    return ERROR;
  }

  return OK;
}

/*------------------------------------------------------------------------
 *
 * velodyneExecTaskEnd  --  Termination codel
 *
 * Description: just closes the driver -- if it has been opened
 *
 * Returns:    OK or ERROR
 */

STATUS
velodyneExecTaskEnd(void)
{
  free (threeDImage);
  free (targetRelativePosition);
  free (targetAbsolutePosition);
  velodyne_close (device);
  return OK;
}

/*------------------------------------------------------------------------------
 * AddLineToFOVFilter - Main
 *
 * Description: adds a line to the FOV filter
 *
 * Reports: OK
 *          S_velodyne_BAD_FOV_PARAMS
 *
 * Returns: OK ETHER
 */
ACTIVITY_EVENT
velodyneAddLineToFOVFilterMain(velodyneFOVParams* fovParams, int* report)
{
  int line;
  int result;
  for (line = 0; line < VELODYNE_3D_IMAGE_HEIGHT; line++)
    if (imageParams.lineOrder[line] == fovParams->line)
      break;
  result = velodyne_add_laser_to_fov_filter(device->model, line);
  if (result == VELODYNE_ERR_FOV) {
    fprintf (stderr, "velodyne: bad line number\n");
    *report = S_velodyne_BAD_FOV_PARAMS;
    return ETHER;
  }
  return OK;
}

/*------------------------------------------------------------------------
 * OneShot
 *
 * Description: grabs scans of data from the device
 *
 * Reports:      OK
 *              S_velodyne_DEVICE_NOT_OPEN
 *              S_velodyne_ACQUISITION_NOT_STARTED
 *              S_velodyne_PARTIAL_SCAN_ACQUISITION
 *              S_velodyne_CAN_NOT_UPDATE_3D_POSTER
 *              S_velodyne_CAN_NOT_UPDATE_TARGET_POSTER
 *		S_velodyne_CAN_NOT_CONVERT_T3D
 *		S_velodyne_FAILED_TO_ANALYSE_TARGET
 */

/* velodyneOneShotStart  -  codel START of OneShot
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
velodyneOneShotStart(velodyneOneShotParams* oneShotParams, int *report)
{
  if (SDI_F->running == 0) {
    fprintf (stderr, "velodyne: the acquisition has not been started\n");
    *report = S_velodyne_ACQUISITION_NOT_STARTED;
    return ETHER;
  }
  if (device == NULL) {
	fprintf (stderr, "velodyne: the device is not open\n");
    *report = S_velodyne_DEVICE_NOT_OPEN;
    return ETHER;
  }
  /* Clean the target structure */
  if (SDI_F->targetDetectionParams.detectionOn == VELODYNE_TRUE) {
    memset (&target, 0, sizeof(velodyneTarget));
  }

  /* Cleans the 3D image */
  threeDImage->width = 0;
  memset (&threeDImage->position, 0, sizeof (POM_SENSOR_POS));
  memset (threeDImage->points, 0, VELODYNE_3D_IMAGE_HEIGHT
                                  * VELODYNE_3D_IMAGE_WIDTH
                                  * sizeof (velodyne3DPoint));
  return EXEC;
}

/* velodyneOneShotMain  -  codel EXEC of OneShot
   Returns:  EXEC END ETHER FAIL ZOMBIE */

ACTIVITY_EVENT
velodyneOneShotMain(velodyneOneShotParams* oneShotParams, int *report)
{
  ACTIVITY_EVENT grabError;
  /* struct timeval tv_start, tv_end; */
  /* Makes sure that velodyneGrabOneShot is accessed in an atomic way */

  task.needOneShot = true;
  pthread_mutex_lock (&task.mtx);
  grabError = velodyneGrabOneShot (&task.buffer,
                                   oneShotParams,
                                   threeDImage,
                                   &imageParams,
                                   &(SDI_F->targetDetectionParams),
                                   &(SDI_F->targetParams),
                                   &target);
  pthread_cond_signal (&task.cond);
  task.needOneShot = false;
  pthread_mutex_unlock (&task.mtx);
  return grabError;
}

/* velodyneOneShotEnd  -  codel END of OneShot
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
velodyneOneShotEnd(velodyneOneShotParams* oneShotParams, int *report)
{
  T3D		sensorToMain;
  T3D		mainToOrigin;
  float		x, y;
  double*	matrix;
  ACTIVITY_EVENT	targetAnalysis;

  /* Asses whether the target has been detected or not */
  if (SDI_F->targetDetectionParams.detectionOn == VELODYNE_TRUE) {
    if (velodyne_analyzeTarget (&(SDI_F->targetParams), 
				&(SDI_F->targetDetectionParams), 
				&target) != END) {
      fprintf (stderr, "velodyne: velodyne_analyzeTarget failed\n");
	*report = S_velodyne_FAILED_TO_ANALYSE_TARGET;
	return FAIL;
    }

    /* Fill the target poster */
    if (target.status != VELODYNE_TARGET_DETECTED) {
      targetAbsolutePosition->numRef++;
      targetAbsolutePosition->nbPts = 0;
      targetRelativePosition->numRef++;
      targetRelativePosition->nbPts = 0;
    }
    else {
      /* Remove the target points from the 3D data poster if required */
      if (SDI_F->targetDetectionParams.eraseTargetPoints == VELODYNE_TRUE)
	velodyne_eraseTargetPoints (threeDImage, &target);

      /* Update the target relative position */
      targetRelativePosition->numRef++;
      targetRelativePosition->nbPts = 1;
      t3dInit(&sensorToMain, T3D_BRYAN, T3D_ALLOW_CONVERSION);
      sensorToMain.euler.euler[0] = threeDImage->position.sensorToMain.euler.yaw;
      sensorToMain.euler.euler[1] = threeDImage->position.sensorToMain.euler.pitch;
      sensorToMain.euler.euler[2] = threeDImage->position.sensorToMain.euler.roll;
      sensorToMain.euler.euler[3] = threeDImage->position.sensorToMain.euler.x;
      sensorToMain.euler.euler[4] = threeDImage->position.sensorToMain.euler.y;
      sensorToMain.euler.euler[5] = threeDImage->position.sensorToMain.euler.z;
      if (!t3dConvertTo(T3D_MATRIX, &sensorToMain)) {
	fprintf (stderr, "velodyne: velodyneOneShotEnd : can't convert t3d\n");
	*report = S_velodyne_CAN_NOT_CONVERT_T3D;
	return FAIL;
      }
      x = target.cog.coordinates[0];
      y = target.cog.coordinates[1];
      matrix = sensorToMain.matrix.matrix;
      targetRelativePosition->points[0].x = 
	x * *(matrix) + y * *(matrix+1) + *(matrix+3);
      targetRelativePosition->points[0].y = 
	x * *(matrix+4) + y * *(matrix+5) + *(matrix+7);

      /* Update the absolute position */
      targetAbsolutePosition->numRef++;
      targetAbsolutePosition->nbPts = 1;
      t3dInit(&mainToOrigin, T3D_BRYAN, T3D_ALLOW_CONVERSION);
      mainToOrigin.euler.euler[0] = threeDImage->position.mainToOrigin.euler.yaw;
      mainToOrigin.euler.euler[1] = threeDImage->position.mainToOrigin.euler.pitch;
      mainToOrigin.euler.euler[2] = threeDImage->position.mainToOrigin.euler.roll;
      mainToOrigin.euler.euler[3] = threeDImage->position.mainToOrigin.euler.x;
      mainToOrigin.euler.euler[4] = threeDImage->position.mainToOrigin.euler.y;
      mainToOrigin.euler.euler[5] = threeDImage->position.mainToOrigin.euler.z;
      if (!t3dConvertTo(T3D_MATRIX, &mainToOrigin)) {
	fprintf (stderr, "velodyne: velodyneOneShotEnd : can't convert t3d\n");
	*report = S_velodyne_CAN_NOT_CONVERT_T3D;
	return FAIL;
      }
      x = targetRelativePosition->points[0].x;
      y = targetRelativePosition->points[0].y;
      matrix = mainToOrigin.matrix.matrix;
      targetAbsolutePosition->points[0].x = 
	x * *(matrix) + y * *(matrix+1) + *(matrix+3);
      targetAbsolutePosition->points[0].y = 
	x * *(matrix+4) + y * *(matrix+5) + *(matrix+7);
    }
    
    /* Now writes the target posters */
    if (velodyneTargetAbsolutePositionGENPOS_TRAJ_POINTSPosterWrite 
	(VELODYNE_TARGETABSOLUTEPOSITION_POSTER_ID, targetAbsolutePosition) != OK) {
      *report = S_velodyne_CAN_NOT_UPDATE_TARGET_POSTER;
      return FAIL;
    }
    if (velodyneTargetRelativePositionGENPOS_TRAJ_POINTSPosterWrite 
	(VELODYNE_TARGETRELATIVEPOSITION_POSTER_ID, targetRelativePosition) != OK) {
      *report = S_velodyne_CAN_NOT_UPDATE_TARGET_POSTER;
      return FAIL;
    }

  }

  /* Write the 3D points poster */
  if (velodyneThreeDImagevelodyne3DImagePosterWrite
      (VELODYNE_THREEDIMAGE_POSTER_ID,
       threeDImage) != OK) {
    *report = S_velodyne_CAN_NOT_UPDATE_3D_POSTER;
    return FAIL;
  }
  

  return ETHER;
}


/*------------------------------------------------------------------------
 * SetTargetDetectionParams
 *
 * Description: allocates the memory required to track the target
 *
 * Reports:      OK
 *              S_velodyne_CAN_NOT_ALLOCATE_MEMORY
 */

/* Returns:  ETHER */

ACTIVITY_EVENT
velodyneSetTargetDetectionParams(velodyneTargetDetectionParams* detectionParams,
				int *report)
{
  /* Just sets the target structure to "0" -- whatever happens */
  memset (&target, 0, sizeof(velodyneTarget));
  return ETHER;
}

/*------------------------------------------------------------------------
 * DumpTarget
 *
 * Description: saves the target points in an ASCII file
 *
 * Reports:      OK
 *              S_velodyne_CANT_OPEN_FILE
 */

/* velodyneDumpTargetMain  -  codel EXEC of DumpTarget
   Returns:  EXEC END ETHER FAIL ZOMBIE */

ACTIVITY_EVENT
velodyneDumpTargetMain(velodyneSaveParams* saveParams,
		       int *report)
{
  char	fileName[VELODYNE_MAX_STRLEN];
  FILE*	file;
  int	i;
  /* opens the file */
  sprintf (fileName, "%s.%04d",
	   saveParams->baseName, saveParams->num);
  file = fopen (fileName, "w");
  if (file == NULL) {
    *report = S_velodyne_CANT_OPEN_FILE;
    return FAIL;
  }
  for (i = 0; i < target.detectedTargetPoints; i++) {
    if ((target.targetPoints + i)->status == VELODYNE_GOOD_3DPOINT)
      fprintf (file, "%.2f %.2f %.2f\n", 
	       (target.targetPoints + i)->coordinates[0],
	       (target.targetPoints + i)->coordinates[1],
	       (target.targetPoints + i)->coordinates[2]);
  }
  fclose (file);
  return ETHER;
}

/*------------------------------------------------------------------------
 * Save
 *
 * Description:
 *
 * Reports: OK
 *          S_velodyne_CANNOT_OPEN_FILE
 *          S_velodyne_CANNOT_WRITE_ON_FILE
 */

/* velodyneSaveMain  -  codel MAIN of Save
   Returns: ETHER */
ACTIVITY_EVENT
velodyneSaveMain(int *report)
{
  int i, j;
  char fileName[VELODYNE_MAX_STRLEN];
  FILE* file;
  velodyne3DPoint* vel3DPoint;
  VELODYNE_CARTESIAN_POINT cartPt;

  /* Opens file */
  sprintf (fileName, "%s.%04d",
           SDI_F->saveParams.baseName, SDI_F->saveParams.num);
  file = fopen (fileName, "w");
  if (file == NULL) {
    *report = S_velodyne_CANNOT_OPEN_FILE;
    return ETHER;
  }

  /* Writes the header */
  fprintf (file, VELODYNE_FILE_FORMAT);
  fprintf (file, "\n%d %d %d\n",
           threeDImage->height, threeDImage->width, threeDImage->maxScanWidth);

  /* Writes the points */
  for (i = 0; i < threeDImage->height; i++) {
    for (j = 0; j < threeDImage->width; j++) {
      vel3DPoint = velodyne3DImageAt (threeDImage, i, j);
      cartPt.x = vel3DPoint->coordinates[0];
      cartPt.y = vel3DPoint->coordinates[1];
      cartPt.z = vel3DPoint->coordinates[2];
      cartPt.intensity = (u_int8_t) vel3DPoint->intensity;
      cartPt.status = (VELODYNE_POINT_STATUS) vel3DPoint->status;
      if (fwrite (&cartPt, sizeof (VELODYNE_CARTESIAN_POINT), 1, file) != 1) {
        *report = S_velodyne_CANNOT_WRITE_ON_FILE;
        fclose (file);
        return ETHER;
      }
    }
  }

  /* Closes file */
  fclose (file);

  /* Composes the file name and saves the pom position */
  sprintf (fileName, "%s.pos.%04d",
           SDI_F->saveParams.baseName, SDI_F->saveParams.num);
  pomWriteSensorPos (fileName, &(threeDImage->position));

  SDI_F->saveParams.num += 1;

  return ETHER;
}

/*------------------------------------------------------------------------
 * SaveIm3D
 *
 * Description:
 *
 * Reports: OK
 *          S_velodyne_CANNOT_OPEN_FILE
 *          S_velodyne_CANNOT_ALLOCATE_IMAGE_3D
 */

/* velodyneSaveIm3DMain  -  codel MAIN of SaveIm3D
   Returns: ETHER */
ACTIVITY_EVENT
velodyneSaveIm3DMain(int *report)
{
  int i, j;
  char fileName[VELODYNE_MAX_STRLEN];
  DATA_IM3D im3D;
  DATA_PT3D* pt3D;
  velodyne3DPoint* vel3DPoint;

  /* Allocates the DATA_IM3D */
  if (init_data_im3d (&im3D, threeDImage->height, threeDImage->maxScanWidth,
                      0, 0, 0)
      != ERR_IM3D_OK) {
    *report = S_velodyne_CANNOT_ALLOCATE_IMAGE_3D;
    return ETHER;
  }

  /* Copies the 3D image into the DATA_IM3D */
  for (i = 0; i < threeDImage->height; i++) {
    for (j = 0; j < threeDImage->width; j++) {
      pt3D = data_im3d_pt (&im3D, i, j);
      vel3DPoint = velodyne3DImageAt (threeDImage, i, j);
      if (vel3DPoint->status == VELODYNE_GOOD_3DPOINT) {
        pt3D->state = GOOD_PT3D;
        pt3D->coord_1 = vel3DPoint->coordinates[0];
        pt3D->coord_2 = vel3DPoint->coordinates[1];
        pt3D->coord_3 = vel3DPoint->coordinates[2];
      }
    }
  }

  im3D.header.coordonnees = COORDONNEES_CARTESIENNES;
  im3D.header.capteur = CAPTEUR_LASER1;
  im3D.header.repere = REPERE_CAPTEUR;
  im3D.header.points = POINTS_CALCULES;

  /* Composes the file name and saves the DATA_IM3D */
  sprintf (fileName, "%s.%04d",
           SDI_F->saveIm3DParams.baseName, SDI_F->saveIm3DParams.num);
  if (write_data_im3d_binaire (&im3D, fileName) != ERR_IM3D_OK) {
    *report = S_velodyne_CANNOT_OPEN_FILE;
    return ETHER;
  }

  /* Composes the file name and saves the pom position */
  sprintf (fileName, "%s.pos.%04d",
           SDI_F->saveIm3DParams.baseName, SDI_F->saveIm3DParams.num);
  pomWriteSensorPos (fileName, &(threeDImage->position));

  SDI_F->saveIm3DParams.num += 1;

  /* Frees the im3d */
  empty_data_im3d (&im3D);

  return ETHER;
}
